import { DrupalJsonApiParams } from "drupal-jsonapi-params";
import { GetServerSidePropsContext, GetStaticPathsContext } from "next";
import type { RequireAllOrNone } from "type-fest";
declare type ViewWithFilters<T, F> = {
    id: string;
    items?: T;
    filters?: F;
    nextPage?: number;
    opts: QueryOptsWithPagination<null>;
};
/**
 * @template T The items type for this view.
 * @template F The filters type for the view. Use `null` for no filters.
 *
 * @example
 * type ViewWithFilters = View<DrupalNode[], { from: number, to: number }>
 * type ViewWithoutFilters = View<DrupalTaxonomyTerm[], null>
 */
export declare type View<T, F> = F extends null ? Omit<ViewWithFilters<T, F>, "filters"> : ViewWithFilters<T, F>;
declare type QueryOptsPaginated = RequireAllOrNone<{
    page: number;
    limit: number;
}, "page" | "limit"> & {
    context?: GetServerSidePropsContext;
};
/**
 * Use this type helper to define options for query params.
 *
 * @template Options The type definition for the options. Use null if no additional options.
 *
 * @example
 * type ParamOpts = QueryOpts<{
 *    id: string
 * }>
 *
 * export const params: QueryParams<ParamOpts> = () => {}
 */
export declare type QueryOpts<Options> = Options & {
    context?: GetStaticPathsContext | GetServerSidePropsContext;
};
/**
 * Use this type helper to define **paginated** options for query params.
 *
 * @template Options The type definition for the options. Use null if no additional options.
 *
 * @example
 * type ParamOpts = PaginatedQueryParamsOpts<{ type: string }>
 *
 * export const params: QueryParams<ParamOpts> = () => {}
 */
export declare type QueryOptsWithPagination<Options> = Options extends null ? QueryOptsPaginated : Options & QueryOptsPaginated;
/**
 * Use this type helper to define options for query params.
 *
 * @template Options The type definition for the options. Use null if no additional options.
 *
 * @example
 *
 * type ParamOpts = QueryOpts<{
 *    id: string
 * }>
 *
 * export const params: QueryParams<ParamOpts> = () => {}
 */
export declare type QueryParams<Options extends QueryOpts<null> | QueryOptsWithPagination<null>> = (opts?: Options) => DrupalJsonApiParams;
/**
 * Use this type helper to define options and return type for query data.
 *
 * @template Options The type definition for the options. Use null if no additional options.
 * @template Return The return type for the data loader.
 *
 * @example
 *
 * type ParamOpts = QueryOpts<{
 *    id: string
 * }>
 *
 * type Article = {
 *    id: string
 *    title: string
 * }
 *
 * export const data: QueryData<ParamOpts, Article> = async (): Promise<Article> => {
 *    return {
 *        id: "",
 *        title: "",
 *    }
 * }
 */
export declare type QueryData<Options, Return> = (opts?: Options) => Promise<Return>;
declare type Queries<Q> = Record<keyof Q, {
    params: (opts: unknown) => DrupalJsonApiParams;
    data?: (opts: unknown) => unknown;
}>;
export declare type QueryTypeOf<Q extends Queries<Q>, T extends keyof Q, F extends "params" | "data"> = Q[T][F];
export declare type QueryId<Q = null> = keyof Q;
export declare type QueryParamsOpts<Q extends Queries<Q>, T extends keyof Q> = Parameters<QueryTypeOf<Q, T, "params">>[0];
export declare type QueryDataOpts<Q extends Queries<Q>, T extends keyof Q> = Parameters<Q[T]["data"]>[0];
export declare type QueryDataReturn<Q extends Queries<Q>, T extends keyof Q> = ReturnType<QueryTypeOf<Q, T, "data">>;
export declare function withPagination(params: DrupalJsonApiParams, opts: QueryOptsWithPagination<unknown>): DrupalJsonApiParams;
export declare function massageRouteQuery(query: any): any;
export declare function createQueries<Q extends Readonly<Queries<Q>>>(queries: Q): {
    getDataIds: <T extends NonNullable<{ [Key in keyof Q]: Q[Key] extends {
        data: any;
    } ? Key : never; }[keyof Q]>>() => T[];
    getIds: <T_1 extends keyof Q>() => T_1[];
    getParams: <T_2 extends keyof Q, O extends QueryParamsOpts<Q, T_2>>(id?: T_2, opts?: O) => DrupalJsonApiParams;
    getData: <T_3 extends NonNullable<{ [Key_1 in keyof Q]: Q[Key_1] extends {
        data: any;
    } ? Key_1 : never; }[keyof Q]>, O_1 extends QueryDataOpts<Q, T_3>>(id: T_3, opts?: O_1) => Promise<ReturnType<Q[T_3]["data"]>>;
};
export {};
